@@ -25,8 +25,7 @@ class PureXArrayDynamic(BaseDynamic):
             for circular_node_id in nodesInCyclic:
                 circular_node = node.model.getNode(circular_node_id)
                 if not circular_node is None:
-                    circular_node.sendStartCalcNode(
-                        fromDynamic=node.identifier)
+                    node.model.sendStartCalcNode(circular_node_id, fromDynamic=node.identifier)
 
         # create nodes array
         cyclicNodes = []
@@ -89,11 +88,11 @@ class PureXArrayDynamic(BaseDynamic):
             _id = _node["node"].identifier
 
             cyclicDic[_id] = evaluate(
-                _node["initialize"], returnEvaluateTime=False)
+                _node["initialize"], returnEvaluateTime=False, fromDynamic=True)
 
             if not initialValues is None and _id in initialValues:
                 initial_result = evaluate(
-                    initialValues[_id], returnEvaluateTime=False)
+                    initialValues[_id], returnEvaluateTime=False, fromDynamic=True)
                 cyclicDic[_id] = cyclicDic[_id] + initial_result
             # check align
             if cyclicDic[_id].dims[0] != dynamicIndex.name:
@@ -164,9 +163,9 @@ class PureXArrayDynamic(BaseDynamic):
                     try:
                         # use initial values
                         _resultNode = evaluate(
-                            _node["loopDefinition"], cyclicParams, False)
+                            _node["loopDefinition"], cyclicParams, False, fromDynamic=True)
                         _initialValues = evaluate(
-                            initialValues[_id], returnEvaluateTime=False)
+                            initialValues[_id], returnEvaluateTime=False, fromDynamic=True)
                     except Exception as ex:
                         raise ValueError(
                             f"Node '{_id}' failed during dynamic evaluation. Error: {ex}")
@@ -186,7 +185,7 @@ class PureXArrayDynamic(BaseDynamic):
                     try:
                         # dont use use initial values
                         _resultNode = evaluate(
-                            _node["loopDefinition"], cyclicParams, False)
+                            _node["loopDefinition"], cyclicParams, False, fromDynamic=True)
                     except Exception as ex:
                         raise ValueError(
                             f"Node '{_id}' failed during dynamic evaluation. Error: {ex}")
@@ -238,7 +237,7 @@ class PureXArrayDynamic(BaseDynamic):
             for circular_node_id in nodesInCyclic:
                 circular_node = node.model.getNode(circular_node_id)
                 if not circular_node is None:
-                    circular_node.sendEndCalcNode(fromDynamic=node.identifier)
+                    node.model.sendEndCalcNode(circular_node_id, fromDynamic=node.identifier)
 
         evaluate = None
         model = None
@@ -388,6 +387,9 @@ class PureXArrayDynamic(BaseDynamic):
         destination_array = destination
         new_values_array = new_values
 
+        destination_array = self.reshapeDataArrays(
+            dynamic_index_name, destination_array, list(new_values_array.dims))
+
         try:
             destination_array.loc[loc_dic] = new_values_array.values
         except Exception as ex:
@@ -419,3 +421,24 @@ class PureXArrayDynamic(BaseDynamic):
                     *list_dims, transpose_coords=True).values
 
         return destination_array
+
+    def reshapeDataArrays(
+        self,
+        dynamic_index_name: str,
+        destination: xr.DataArray,
+        new_values_dims: list
+    ) -> xr.DataArray:
+        """Reshape dimensions and coords to avoid problems when
+           there are 2 or more dims with same number of elements
+        """
+        destination_array = destination
+
+        destination_dims = list(destination_array.dims)
+        destination_dims.remove(dynamic_index_name)
+        if len(destination_dims) == len(new_values_dims) and destination_dims != new_values_dims:
+            # This should happen only after first assignment; afterwards, it will have the correct shape
+            reordered_dims = [dynamic_index_name] + new_values_dims
+            destination_array = destination_array.transpose(
+                *reordered_dims).copy()  # copying to avoid view error
+
+        return destination_array
